@@ -14,13 +14,21 @@
 # under the License.
 
 import argparse
+import datetime
+import dateutil
 import logging
 import os
 import Queue
+import re
+import subprocess
 import sys
+import textwrap
 import threading
+import urlparse
+import warnings
 import webbrowser
 
+import sqlalchemy.exc
 import urwid
 
 from gertty import db
@@ -31,6 +39,7 @@ from gertty import mywid
 from gertty import palette
 from gertty import sync
 from gertty import search
+from gertty import requestsexceptions
 from gertty.view import change_list as view_change_list
 from gertty.view import project_list as view_project_list
 from gertty.view import change as view_change
@@ -58,8 +67,10 @@ class StatusHeader(urwid.WidgetWrap):
         self.error_widget = urwid.Text('')
         self.offline_widget = urwid.Text('')
         self.sync_widget = urwid.Text(u'Sync: 0')
+        self.held_widget = urwid.Text(u'')
         self._w.contents.append((self.title_widget, ('pack', None, False)))
         self._w.contents.append((urwid.Text(u''), ('weight', 1, False)))
+        self._w.contents.append((self.held_widget, ('pack', None, False)))
         self._w.contents.append((self.error_widget, ('pack', None, False)))
         self._w.contents.append((self.offline_widget, ('pack', None, False)))
         self._w.contents.append((self.sync_widget, ('pack', None, False)))
@@ -67,18 +78,23 @@ class StatusHeader(urwid.WidgetWrap):
         self.offline = None
         self.title = None
         self.sync = None
+        self.held = None
         self._error = False
         self._offline = False
         self._title = ''
         self._sync = 0
+        self._held = 0
+        self.held_key = self.app.config.keymap.formatKeys(keymap.LIST_HELD)
 
-    def update(self, title=None, error=None, offline=None, refresh=True):
+    def update(self, title=None, error=None, offline=None, refresh=True, held=None):
         if title is not None:
             self.title = title
         if error is not None:
             self.error = error
         if offline is not None:
             self.offline = offline
+        if held is not None:
+            self.held = held
         self.sync = self.app.sync.queue.qsize()
         if refresh:
             self.refresh()
@@ -87,6 +103,12 @@ class StatusHeader(urwid.WidgetWrap):
         if self._title != self.title:
             self._title = self.title
             self.title_widget.set_text(self._title)
+        if self._held != self.held:
+            self._held = self.held
+            if self._held:
+                self.held_widget.set_text(('error', u'Held: %s (%s)' % (self._held, self.held_key)))
+            else:
+                self.held_widget.set_text(u'')
         if self._error != self.error:
             self._error = self.error
             if self._error:
@@ -128,7 +150,34 @@ class SearchDialog(mywid.ButtonDialog):
             return None
         return r
 
+# From: cpython/file/2.7/Lib/webbrowser.py with modification to
+# redirect stdin/out/err.
+class BackgroundBrowser(webbrowser.GenericBrowser):
+    """Class for all browsers which are to be started in the
+       background."""
+
+    def open(self, url, new=0, autoraise=True):
+        cmdline = [self.name] + [arg.replace("%s", url)
+                                 for arg in self.args]
+        inout = file(os.devnull, "r+")
+        try:
+            if sys.platform[:3] == 'win':
+                p = subprocess.Popen(cmdline)
+            else:
+                setsid = getattr(os, 'setsid', None)
+                if not setsid:
+                    setsid = getattr(os, 'setpgrp', None)
+                p = subprocess.Popen(cmdline, close_fds=True,
+                                     stdin=inout, stdout=inout,
+                                     stderr=inout, preexec_fn=setsid)
+            return (p.poll() is None)
+        except OSError:
+            return False
+
 class App(object):
+    simple_change_search = re.compile('^(\d+|I[a-fA-F0-9]{40})$')
+    url_change_search = re.compile('^https?://.*$')
+
     def __init__(self, server=None, palette='default', keymap='default',
                  debug=False, verbose=False, disable_sync=False,
                  fetch_missing_refs=False, path=config.DEFAULT_CONFIG_PATH):
@@ -157,6 +206,8 @@ class App(object):
         self.log = logging.getLogger('gertty.App')
         self.log.debug("Starting")
 
+        webbrowser.register('xdg-open', None, BackgroundBrowser("xdg-open"))
+
         self.fetch_missing_refs = fetch_missing_refs
         self.config.keymap.updateCommandMap()
         self.search = search.SearchCompiler(self)
@@ -168,9 +219,17 @@ class App(object):
         self.header = urwid.AttrMap(self.status, 'header')
         screen = view_project_list.ProjectListView(self)
         self.status.update(title=screen.title)
+        self.updateStatusQueries()
         self.loop = urwid.MainLoop(screen, palette=self.config.palette.getPalette(),
                                    unhandled_input=self.unhandledInput)
 
+        self.sync_pipe = self.loop.watch_pipe(self.refresh)
+        self.error_queue = Queue.Queue()
+        self.error_pipe = self.loop.watch_pipe(self._errorPipeInput)
+        self.logged_warnings = set()
+
+        warnings.showwarning = self._showWarning
+
         has_subscribed_projects = False
         with self.db.getSession() as session:
             if session.getProjects(subscribed=True):
@@ -178,7 +237,6 @@ class App(object):
         if not has_subscribed_projects:
             self.welcome()
 
-        self.sync_pipe = self.loop.watch_pipe(self.refresh)
         self.loop.screen.tty_signal_keys(start='undefined', stop='undefined')
         #self.loop.screen.set_terminal_properties(colors=88)
         if not disable_sync:
@@ -240,20 +298,30 @@ class App(object):
             self.loop.widget = widget
 
     def refresh(self, data=None, force=False):
-        self.status.refresh()
         widget = self.loop.widget
         while isinstance(widget, urwid.Overlay):
             widget = widget.contents[0][0]
         interested = force
+        invalidate = False
         try:
             while True:
                 event = self.sync.result_queue.get(0)
                 if widget.interested(event):
                     interested = True
+                if hasattr(event, 'held_changed') and event.held_changed:
+                    invalidate = True
         except Queue.Empty:
             pass
         if interested:
             widget.refresh()
+        if invalidate:
+            self.updateStatusQueries()
+        self.status.refresh()
+
+    def updateStatusQueries(self):
+        with self.db.getSession() as session:
+            held = len(session.getHeld())
+            self.status.update(held=held)
 
     def popup(self, widget,
               relative_width=50, relative_height=25,
@@ -351,6 +419,10 @@ class App(object):
                 changes = session.getChanges(query)
             except gertty.search.SearchSyntaxError as e:
                 return self.error(e.message)
+            except sqlalchemy.exc.OperationalError as e:
+                return self.error(e.message)
+            except Exception as e:
+                return self.error(str(e))
             change_key = None
             if len(changes) == 1:
                 change_key = changes[0].key
@@ -373,33 +445,80 @@ class App(object):
 
     def _searchDialog(self, dialog):
         self.backScreen()
-        query = dialog.entry.edit_text
-        try:
-            query = 'change:%s' % int(query)
-        except ValueError:
-            pass
+        query = dialog.entry.edit_text.strip()
+        if self.simple_change_search.match(query):
+            query = 'change:%s' % query
+        else:
+            result = self.parseInternalURL(query)
+            if result is not None:
+                return self.openInternalURL(result)
         self.doSearch(query)
 
-    def error(self, message):
-        dialog = mywid.MessageDialog('Error', message)
+    trailing_filename_re = re.compile('.*(,[a-z]+)')
+    def parseInternalURL(self, url):
+        if not self.url_change_search.match(url):
+            return None
+        result = urlparse.urlparse(url)
+        if result.netloc != self.config.hostname:
+            return None
+        change = patchset = filename = None
+        path = [x for x in result.path.split('/') if x]
+        if path:
+            change = path[0]
+        else:
+            path = [x for x in result.fragment.split('/') if x]
+            if path[0] == 'c':
+                path.pop(0)
+            while path:
+                if not change:
+                    change = path.pop(0)
+                    continue
+                if not patchset:
+                    patchset = path.pop(0)
+                    continue
+                if not filename:
+                    filename = '/'.join(path)
+                    m = self.trailing_filename_re.match(filename)
+                    if m:
+                        filename = filename[:0-len(m.group(1))]
+                    path = None
+        return (change, patchset, filename)
+
+    def openInternalURL(self, result):
+        (change, patchset, filename) = result
+        # TODO: support deep-linking to a filename
+        self.doSearch('change:%s' % change)
+
+    def error(self, message, title='Error'):
+        dialog = mywid.MessageDialog(title, message)
         urwid.connect_signal(dialog, 'close',
                              lambda button: self.backScreen())
-        self.popup(dialog, min_height=4)
+
+        cols, rows = self.loop.screen.get_cols_rows()
+        cols = int(cols*.5)
+        lines = textwrap.wrap(message, cols)
+        min_height = max(4, len(lines)+4)
+
+        self.popup(dialog, min_height=min_height)
         return None
 
     def unhandledInput(self, key):
         commands = self.config.keymap.getCommands(key)
         if keymap.PREV_SCREEN in commands:
             self.backScreen()
+        elif keymap.TOP_SCREEN in commands:
+            self.clearHistory()
+            self.refresh(force=True)
         elif keymap.HELP in commands:
             self.help()
         elif keymap.QUIT in commands:
             self.quit()
         elif keymap.CHANGE_SEARCH in commands:
             self.searchDialog()
+        elif keymap.LIST_HELD in commands:
+            self.doSearch("status:open is:held")
         elif key in self.config.dashboards:
             d = self.config.dashboards[key]
-            self.clearHistory()
             view = view_change_list.ChangeListView(self, d['query'], d['name'])
             self.changeScreen(view)
 
@@ -414,8 +533,131 @@ class App(object):
         self.log.debug("Open URL %s" % url)
         webbrowser.open_new_tab(url)
 
+    def time(self, dt):
+        utc = dt.replace(tzinfo=dateutil.tz.tzutc())
+        if self.config.utc:
+            return utc
+        local = utc.astimezone(dateutil.tz.tzlocal())
+        return local
+
+    def _errorPipeInput(self, data=None):
+        (title, message) = self.error_queue.get()
+        self.error(message, title=title)
+
+    def _showWarning(self, message, category, filename, lineno,
+                     file=None, line=None):
+        # Don't display repeat warnings
+        if str(message) in self.logged_warnings:
+            return
+        m = warnings.formatwarning(message, category, filename, lineno, line)
+        self.log.warning(m)
+        self.logged_warnings.add(str(message))
+        # Log this warning, but never display it to the user; it is
+        # nearly un-actionable.
+        if category == requestsexceptions.InsecurePlatformWarning:
+            return
+        # Disable InsecureRequestWarning when certificate validation is disabled
+        if not self.config.verify_ssl:
+            if category == requestsexceptions.InsecureRequestWarning:
+                return
+        self.error_queue.put(('Warning', m))
+        os.write(self.error_pipe, 'error\n')
+
+    def toggleHeldChange(self, change_key):
+        with self.db.getSession() as session:
+            change = session.getChange(change_key)
+            change.held = not change.held
+            ret = change.held
+            if not change.held:
+                for r in change.revisions:
+                    for m in change.messages:
+                        if m.pending:
+                            self.sync.submitTask(
+                                sync.UploadReviewTask(m.key, sync.HIGH_PRIORITY))
+        self.updateStatusQueries()
+        return ret
+
+    def localCheckoutCommit(self, project_name, commit_sha):
+        repo = self.getRepo(project_name)
+        try:
+            repo.checkout(commit_sha)
+            dialog = mywid.MessageDialog('Checkout', 'Change checked out in %s' % repo.path)
+            min_height=8
+        except gitrepo.GitCheckoutError as e:
+            dialog = mywid.MessageDialog('Error', e.msg)
+            min_height=12
+        urwid.connect_signal(dialog, 'close',
+            lambda button: self.backScreen())
+        self.popup(dialog, min_height=min_height)
+
+    def localCherryPickCommit(self, project_name, commit_sha):
+        repo = self.getRepo(project_name)
+        try:
+            repo.cherryPick(commit_sha)
+            dialog = mywid.MessageDialog('Cherry-Pick', 'Change cherry-picked in %s' % repo.path)
+            min_height=8
+        except gitrepo.GitCheckoutError as e:
+            dialog = mywid.MessageDialog('Error', e.msg)
+            min_height=12
+        urwid.connect_signal(dialog, 'close',
+            lambda button: self.backScreen())
+        self.popup(dialog, min_height=min_height)
+
+    def saveReviews(self, revision_keys, approvals, message, upload, submit):
+        message_keys = []
+        with self.db.getSession() as session:
+            account = session.getAccountByUsername(self.config.username)
+            for revision_key in revision_keys:
+                k = self._saveReview(session, account, revision_key,
+                                     approvals, message, upload, submit)
+                if k:
+                    message_keys.append(k)
+        return message_keys
+
+    def _saveReview(self, session, account, revision_key,
+                    approvals, message, upload, submit):
+        message_key = None
+        revision = session.getRevision(revision_key)
+        change = revision.change
+        draft_approvals = {}
+        for approval in change.draft_approvals:
+            draft_approvals[approval.category] = approval
+
+        categories = set()
+        for label in change.permitted_labels:
+            categories.add(label.category)
+        for category in categories:
+            value = approvals.get(category, 0)
+            approval = draft_approvals.get(category)
+            if not approval:
+                approval = change.createApproval(account, category, 0, draft=True)
+                draft_approvals[category] = approval
+            approval.value = value
+        draft_message = revision.getPendingMessage()
+        if not draft_message:
+            draft_message = revision.getDraftMessage()
+        if not draft_message:
+            if message or upload:
+                draft_message = revision.createMessage(None, account,
+                                                       datetime.datetime.utcnow(),
+                                                       '', draft=True)
+        if draft_message:
+            draft_message.created = datetime.datetime.utcnow()
+            draft_message.message = message
+            draft_message.pending = upload
+            message_key = draft_message.key
+        if upload:
+            change.reviewed = True
+        if submit:
+            change.status = 'SUBMITTED'
+            change.pending_status = True
+            change.pending_status_message = None
+        return message_key
+
+
+
 def version():
-    return "Gertty version: %s" % gertty.version.version_info.version_string()
+    return "Gertty version: %s" % gertty.version.version_info.release_string()
 
 class PrintKeymapAction(argparse.Action):
     def __call__(self, parser, namespace, values, option_string=None):
