@@ -21,6 +21,7 @@ try:
     import ordereddict
 except:
     pass
+import urlparse
 import yaml
 
 import voluptuous as v
@@ -93,6 +94,9 @@ class ConfigSchema(object):
 
     hide_comments = [hide_comment]
 
+    change_list_options = {'sort-by': v.Any('number', 'updated'),
+                           'reverse': bool}
+
     keymap = {v.Required('name'): str,
               v.Match('(?!name)'): v.Any([str], str)}
 
@@ -111,6 +115,9 @@ class ConfigSchema(object):
                            'diff-view': str,
                            'hide-comments': self.hide_comments,
                            'thread-changes': bool,
+                           'display-times-in-utc': bool,
+                           'change-list-options': self.change_list_options,
+                           'expire-age': str,
                            })
         return schema
 
@@ -132,6 +139,8 @@ class Config(object):
         if not url.endswith('/'):
             url += '/'
         self.url = url
+        result = urlparse.urlparse(url)
+        self.hostname = result.netloc
         self.username = server['username']
         self.password = server.get('password')
         if self.password is None:
@@ -212,6 +221,14 @@ class Config(object):
             self.hide_comments.append(re.compile(h['author']))
 
         self.thread_changes = self.config.get('thread-changes', True)
+        self.utc = self.config.get('display-times-in-utc', False)
+
+        change_list_options = self.config.get('change-list-options', {})
+        self.change_list_options = {
+            'sort-by': change_list_options.get('sort-by', 'number'),
+            'reverse': change_list_options.get('reverse', False)}
+
+        self.expire_age = self.config.get('expire-age', '2 months')
 
     def getServer(self, name=None):
         for server in self.config['servers']:
