@@ -13,6 +13,7 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
+import re
 import time
 import logging
 import threading
@@ -20,9 +21,9 @@ import threading
 import alembic
 import alembic.config
 import sqlalchemy
-from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String, Boolean, DateTime, Text, select, func
+from sqlalchemy import create_engine, MetaData, Table, Column, Integer, String, Boolean, DateTime, Text
 from sqlalchemy.schema import ForeignKey
-from sqlalchemy.orm import mapper, sessionmaker, relationship, column_property, scoped_session
+from sqlalchemy.orm import mapper, sessionmaker, relationship, scoped_session
 from sqlalchemy.orm.session import Session
 from sqlalchemy.sql import exists
 from sqlalchemy.sql.expression import and_
@@ -58,8 +59,11 @@ change_table = Table(
     Column('status', String(16), index=True, nullable=False),
     Column('hidden', Boolean, index=True, nullable=False),
     Column('reviewed', Boolean, index=True, nullable=False),
+    Column('starred', Boolean, index=True, nullable=False),
+    Column('held', Boolean, index=True, nullable=False),
     Column('pending_rebase', Boolean, index=True, nullable=False),
     Column('pending_topic', Boolean, index=True, nullable=False),
+    Column('pending_starred', Boolean, index=True, nullable=False),
     Column('pending_status', Boolean, index=True, nullable=False),
     Column('pending_status_message', Text),
     )
@@ -69,8 +73,9 @@ revision_table = Table(
     Column('change_key', Integer, ForeignKey("change.key"), index=True),
     Column('number', Integer, index=True, nullable=False),
     Column('message', Text, nullable=False),
-    Column('commit', String(255), nullable=False),
-    Column('parent', String(255), nullable=False),
+    Column('commit', String(255), index=True, nullable=False),
+    Column('parent', String(255), index=True, nullable=False),
+    # TODO: fetch_ref, fetch_auth are unused; remove
     Column('fetch_auth', Boolean, nullable=False),
     Column('fetch_ref', String(255), nullable=False),
     Column('pending_message', Boolean, index=True, nullable=False),
@@ -90,12 +95,11 @@ message_table = Table(
 comment_table = Table(
     'comment', metadata,
     Column('key', Integer, primary_key=True),
-    Column('revision_key', Integer, ForeignKey("revision.key"), index=True),
+    Column('file_key', Integer, ForeignKey("file.key"), index=True),
     Column('account_key', Integer, ForeignKey("account.key"), index=True),
     Column('id', String(255), index=True), #, unique=True, nullable=False),
     Column('in_reply_to', String(255)),
     Column('created', DateTime, index=True, nullable=False),
-    Column('file', Text, nullable=False),
     Column('parent', Boolean, nullable=False),
     Column('line', Integer),
     Column('message', Text, nullable=False),
@@ -142,6 +146,23 @@ pending_cherry_pick_table = Table(
     Column('branch', String(255), nullable=False),
     Column('message', Text, nullable=False),
     )
+sync_query_table = Table(
+    'sync_query', metadata,
+    Column('key', Integer, primary_key=True),
+    Column('name', String(255), index=True, unique=True, nullable=False),
+    Column('updated', DateTime, index=True),
+    )
+file_table = Table(
+    'file', metadata,
+    Column('key', Integer, primary_key=True),
+    Column('revision_key', Integer, ForeignKey("revision.key"), index=True),
+    Column('path', Text, nullable=False, index=True),
+    Column('old_path', Text, index=True),
+    Column('inserted', Integer),
+    Column('deleted', Integer),
+    Column('status', String(1), nullable=False),
+    )
+
 
 class Account(object):
     def __init__(self, id, name=None, username=None, email=None):
@@ -180,11 +201,12 @@ class Branch(object):
         self.name = name
 
 class Change(object):
-    def __init__(self, project, id, owner, number, branch,
-                 change_id, subject, created, updated, status,
-                 topic=None, hidden=False, reviewed=False,
+    def __init__(self, project, id, owner, number, branch, change_id,
+                 subject, created, updated, status, topic=None,
+                 hidden=False, reviewed=False, starred=False, held=False,
                  pending_rebase=False, pending_topic=False,
-                 pending_status=False, pending_status_message=None):
+                 pending_starred=False, pending_status=False,
+                 pending_status_message=None):
         self.project_key = project.key
         self.account_key = owner.key
         self.id = id
@@ -198,8 +220,11 @@ class Change(object):
         self.status = status
         self.hidden = hidden
         self.reviewed = reviewed
+        self.starred = starred
+        self.held = held
         self.pending_rebase = pending_rebase
         self.pending_topic = pending_topic
+        self.pending_starred = pending_starred
         self.pending_status = pending_status
         self.pending_status_message = pending_status_message
 
@@ -217,6 +242,8 @@ class Change(object):
         cat_max = {}
         cat_value = {}
         for approval in self.approvals:
+            if approval.draft:
+                continue
             cur_min = cat_min.get(approval.category, 0)
             cur_max = cat_max.get(approval.category, 0)
             cur_min = min(approval.value, cur_min)
@@ -231,6 +258,22 @@ class Change(object):
             cat_value[approval.category] = cur_value
         self._approval_cache = cat_value
 
+    def getMinMaxPermittedForCategory(self, category):
+        if not hasattr(self, '_permitted_cache'):
+            self._updatePermittedCache()
+        return self._permitted_cache.get(category, (0,0))
+
+    def _updatePermittedCache(self):
+        cache = {}
+        for label in self.labels:
+            if label.category not in cache:
+                cache[label.category] = [0, 0]
+            if label.value > cache[label.category][1]:
+                cache[label.category][1] = label.value
+            if label.value < cache[label.category][0]:
+                cache[label.category][0] = label.value
+        self._permitted_cache = cache
+
     def createRevision(self, *args, **kw):
         session = Session.object_session(self)
         args = [self] + list(args)
@@ -303,15 +346,6 @@ class Revision(object):
         session.flush()
         return m
 
-    def createComment(self, *args, **kw):
-        session = Session.object_session(self)
-        args = [self] + list(args)
-        c = Comment(*args, **kw)
-        self.comments.append(c)
-        session.add(c)
-        session.flush()
-        return c
-
     def createPendingCherryPick(self, *args, **kw):
         session = Session.object_session(self)
         args = [self] + list(args)
@@ -321,6 +355,24 @@ class Revision(object):
         session.flush()
         return c
 
+    def createFile(self, *args, **kw):
+        session = Session.object_session(self)
+        args = [self] + list(args)
+        f = File(*args, **kw)
+        self.files.append(f)
+        session.add(f)
+        session.flush()
+        if hasattr(self, '_file_cache'):
+            self._file_cache[f.path] = f
+        return f
+
+    def getFile(self, path):
+        if not hasattr(self, '_file_cache'):
+            self._file_cache = {}
+            for f in self.files:
+                self._file_cache[f.path] = f
+        return self._file_cache.get(path, None)
+
     def getPendingMessage(self):
         for m in self.messages:
             if m.pending:
@@ -357,13 +409,12 @@ class Message(object):
         return author_name
 
 class Comment(object):
-    def __init__(self, revision, id, author, in_reply_to, created, file, parent, line, message, draft=False):
-        self.revision_key = revision.key
+    def __init__(self, file, id, author, in_reply_to, created, parent, line, message, draft=False):
+        self.file_key = file.key
         self.account_key = author.key
         self.id = id
         self.in_reply_to = in_reply_to
         self.created = created
-        self.file = file
         self.parent = parent
         self.line = line
         self.message = message
@@ -396,12 +447,71 @@ class PendingCherryPick(object):
         self.branch = branch
         self.message = message
 
+class SyncQuery(object):
+    def __init__(self, name):
+        self.name = name
+
+class File(object):
+    STATUS_ADDED = 'A'
+    STATUS_DELETED = 'D'
+    STATUS_RENAMED = 'R'
+    STATUS_COPIED = 'C'
+    STATUS_REWRITTEN = 'W'
+    STATUS_MODIFIED = 'M'
+
+    def __init__(self, revision, path, status, old_path=None,
+                 inserted=None, deleted=None):
+        self.revision_key = revision.key
+        self.path = path
+        self.status = status
+        self.old_path = old_path
+        self.inserted = inserted
+        self.deleted = deleted
+
+    @property
+    def display_path(self):
+        if not self.old_path:
+            return self.path
+        pre = []
+        post = []
+        for start in range(min(len(self.old_path), len(self.path))):
+            if self.path[start] == self.old_path[start]:
+                pre.append(self.old_path[start])
+            else:
+                break
+        pre = ''.join(pre)
+        for end in range(1, min(len(self.old_path), len(self.path))-1):
+            if self.path[0-end] == self.old_path[0-end]:
+                post.insert(0, self.old_path[0-end])
+            else:
+                break
+        post = ''.join(post)
+        mid = '{%s => %s}' % (self.old_path[start:0-end+1], self.path[start:0-end+1])
+        if pre and post:
+            mid = '{%s => %s}' % (self.old_path[start:0-end+1],
+                                  self.path[start:0-end+1])
+            return pre + mid + post
+        else:
+            return '%s => %s' % (self.old_path, self.path)
+
+    def createComment(self, *args, **kw):
+        session = Session.object_session(self)
+        args = [self] + list(args)
+        c = Comment(*args, **kw)
+        self.comments.append(c)
+        session.add(c)
+        session.flush()
+        return c
+
+
 mapper(Account, account_table)
 mapper(Project, project_table, properties=dict(
         branches=relationship(Branch, backref='project',
-                              order_by=branch_table.c.name),
+                              order_by=branch_table.c.name,
+                              cascade='all, delete-orphan'),
         changes=relationship(Change, backref='project',
-                             order_by=change_table.c.number),
+                             order_by=change_table.c.number,
+                             cascade='all, delete-orphan'),
         unreviewed_changes=relationship(Change,
                                         primaryjoin=and_(project_table.c.key==change_table.c.project_key,
                                                          change_table.c.hidden==False,
@@ -421,17 +531,23 @@ mapper(Branch, branch_table)
 mapper(Change, change_table, properties=dict(
         owner=relationship(Account),
         revisions=relationship(Revision, backref='change',
-                               order_by=revision_table.c.number),
+                               order_by=revision_table.c.number,
+                               cascade='all, delete-orphan'),
         messages=relationship(Message,
                               secondary=revision_table,
-                              order_by=message_table.c.created),
-        labels=relationship(Label, backref='change', order_by=(label_table.c.category,
-                                                               label_table.c.value)),
+                              order_by=message_table.c.created,
+                              viewonly=True),
+        labels=relationship(Label, backref='change',
+                            order_by=(label_table.c.category, label_table.c.value),
+                            cascade='all, delete-orphan'),
         permitted_labels=relationship(PermittedLabel, backref='change',
                                       order_by=(permitted_label_table.c.category,
-                                                permitted_label_table.c.value)),
-        approvals=relationship(Approval, backref='change', order_by=(approval_table.c.category,
-                                                                     approval_table.c.value)),
+                                                permitted_label_table.c.value),
+                                      cascade='all, delete-orphan'),
+        approvals=relationship(Approval, backref='change',
+                               order_by=(approval_table.c.category,
+                                         approval_table.c.value),
+                               cascade='all, delete-orphan'),
         draft_approvals=relationship(Approval,
                                      primaryjoin=and_(change_table.c.key==approval_table.c.change_key,
                                                       approval_table.c.draft==True),
@@ -439,19 +555,27 @@ mapper(Change, change_table, properties=dict(
                                                approval_table.c.value))
         ))
 mapper(Revision, revision_table, properties=dict(
-        messages=relationship(Message, backref='revision'),
-        comments=relationship(Comment, backref='revision',
-                              order_by=(comment_table.c.line,
-                                        comment_table.c.created)),
-        draft_comments=relationship(Comment,
-                                    primaryjoin=and_(revision_table.c.key==comment_table.c.revision_key,
-                                                     comment_table.c.draft==True),
-                                    order_by=(comment_table.c.line,
-                                              comment_table.c.created)),
-        pending_cherry_picks=relationship(PendingCherryPick, backref='revision'),
+        messages=relationship(Message, backref='revision',
+                              cascade='all, delete-orphan'),
+        files=relationship(File, backref='revision',
+                           cascade='all, delete-orphan'),
+        pending_cherry_picks=relationship(PendingCherryPick, backref='revision',
+                                          cascade='all, delete-orphan'),
         ))
 mapper(Message, message_table, properties=dict(
         author=relationship(Account)))
+mapper(File, file_table, properties=dict(
+       comments=relationship(Comment, backref='file',
+                             order_by=(comment_table.c.line,
+                                       comment_table.c.created),
+                             cascade='all, delete-orphan'),
+       draft_comments=relationship(Comment,
+                                   primaryjoin=and_(file_table.c.key==comment_table.c.file_key,
+                                                    comment_table.c.draft==True),
+                                   order_by=(comment_table.c.line,
+                                             comment_table.c.created)),
+       ))
+
 mapper(Comment, comment_table, properties=dict(
         author=relationship(Account)))
 mapper(Label, label_table)
@@ -459,6 +583,16 @@ mapper(PermittedLabel, permitted_label_table)
 mapper(Approval, approval_table, properties=dict(
         reviewer=relationship(Account)))
 mapper(PendingCherryPick, pending_cherry_pick_table)
+mapper(SyncQuery, sync_query_table)
+
+def match(expr, item):
+    if item is None:
+        return False
+    return re.match(expr, item) is not None
+
+@sqlalchemy.event.listens_for(sqlalchemy.engine.Engine, "connect")
+def add_sqlite_match(dbapi_connection, connection_record):
+    dbapi_connection.create_function("matches", 2, match)
 
 class Database(object):
     def __init__(self, app):
@@ -466,15 +600,22 @@ class Database(object):
         self.app = app
         self.engine = create_engine(self.app.config.dburi)
         #metadata.create_all(self.engine)
-        self.migrate()
-        self.session_factory = sessionmaker(bind=self.engine)
+        self.migrate(app)
+        # If we want the objects returned from query() to be usable
+        # outside of the session, we need to expunge them from the session,
+        # and since the DatabaseSession always calls commit() on the session
+        # when the context manager exits, we need to inform the session to
+        # expire objects when it does so.
+        self.session_factory = sessionmaker(bind=self.engine,
+                                            expire_on_commit=False,
+                                            autoflush=False)
         self.session = scoped_session(self.session_factory)
         self.lock = threading.Lock()
 
     def getSession(self):
         return DatabaseSession(self)
 
-    def migrate(self):
+    def migrate(self, app):
         conn = self.engine.connect()
         context = alembic.migration.MigrationContext.configure(conn)
         current_rev = context.get_current_revision()
@@ -485,6 +626,7 @@ class Database(object):
         config = alembic.config.Config()
         config.set_main_option("script_location", "gertty:alembic")
         config.set_main_option("sqlalchemy.url", self.app.config.dburi)
+        config.gertty_app = app
 
         if current_rev is None and has_table:
             self.log.debug('Stamping database as initial revision')
@@ -522,6 +664,9 @@ class DatabaseSession(object):
     def delete(self, obj):
         self.session().delete(obj)
 
+    def vacuum(self):
+        self.session().execute("VACUUM")
+
     def getProjects(self, subscribed=False, unreviewed=False):
         """Retrieve projects.
 
@@ -548,6 +693,12 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
+    def getSyncQueryByName(self, name):
+        try:
+            return self.session().query(SyncQuery).filter_by(name=name).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            return self.createSyncQuery(name)
+
     def getChange(self, key):
         try:
             return self.session().query(Change).filter_by(key=key).one()
@@ -560,6 +711,14 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
+    def getChangeIDs(self, ids):
+        # Returns a set of IDs that exist in the local database matching
+        # the set of supplied IDs. This is used when sync'ing the changesets
+        # locally with the remote changes.
+        if not ids:
+            return set([])
+        return set([r[0] for r in self.session().query(Change.id).filter(Change.id.in_(ids)).all()])
+
     def getChangeByChangeID(self, change_id):
         try:
             return self.session().query(Change).filter_by(change_id=change_id).one()
@@ -578,12 +737,15 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
-    def getChanges(self, query, unreviewed=False):
+    def getChanges(self, query, unreviewed=False, sort_by='number'):
         self.database.log.debug("Search query: %s" % query)
         q = self.session().query(Change).filter(self.search.parse(query))
         if unreviewed:
             q = q.filter(change_table.c.hidden==False, change_table.c.reviewed==False)
-        q = q.order_by(change_table.c.number)
+        if sort_by == 'updated':
+            q = q.order_by(change_table.c.updated)
+        else:
+            q = q.order_by(change_table.c.number)
         self.database.log.debug("Search SQL: %s" % q)
         try:
             return q.all()
@@ -603,8 +765,10 @@ class DatabaseSession(object):
             return None
 
     def getRevisionsByParent(self, parent):
+        if isinstance(parent, basestring):
+            parent = (parent,)
         try:
-            return self.session().query(Revision).filter_by(parent=parent).all()
+            return self.session().query(Revision).filter(Revision.parent.in_(parent)).all()
         except sqlalchemy.orm.exc.NoResultFound:
             return []
 
@@ -614,6 +778,12 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
+    def getFile(self, key):
+        try:
+            return self.session().query(File).filter_by(key=key).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            return None
+
     def getComment(self, key):
         try:
             return self.session().query(Comment).filter_by(key=key).one()
@@ -638,6 +808,9 @@ class DatabaseSession(object):
         except sqlalchemy.orm.exc.NoResultFound:
             return None
 
+    def getHeld(self):
+        return self.session().query(Change).filter_by(held=True).all()
+
     def getPendingMessages(self):
         return self.session().query(Message).filter_by(pending=True).all()
 
@@ -647,6 +820,9 @@ class DatabaseSession(object):
     def getPendingRebases(self):
         return self.session().query(Change).filter_by(pending_rebase=True).all()
 
+    def getPendingStarred(self):
+        return self.session().query(Change).filter_by(pending_starred=True).all()
+
     def getPendingStatusChanges(self):
         return self.session().query(Change).filter_by(pending_status=True).all()
 
@@ -689,3 +865,9 @@ class DatabaseSession(object):
         self.session().add(a)
         self.session().flush()
         return a
+
+    def createSyncQuery(self, *args, **kw):
+        o = SyncQuery(*args, **kw)
+        self.session().add(o)
+        self.session().flush()
+        return o
