@@ -170,6 +170,8 @@ class DiffFile(object):
     def __init__(self):
         self.newname = 'Unknown File'
         self.oldname = 'Unknown File'
+        self.old_empty = False
+        self.new_empty = False
         self.chunks = []
         self.current_chunk = None
         self.old_lineno = 0
@@ -234,6 +236,11 @@ class GitCheckoutError(Exception):
         super(GitCheckoutError, self).__init__(msg)
         self.msg = msg
 
+class GitCloneError(Exception):
+    def __init__(self, msg):
+        super(GitCloneError, self).__init__(msg)
+        self.msg = msg
+
 class Repo(object):
     def __init__(self, url, path):
         self.log = logging.getLogger('gertty.gitrepo')
@@ -241,10 +248,9 @@ class Repo(object):
         self.path = path
         self.differ = difflib.Differ()
         if not os.path.exists(path):
+            if url is None:
+                raise GitCloneError("No URL available for git clone")
             git.Repo.clone_from(self.url, self.path)
-            self.newly_cloned = True
-        else:
-            self.newly_cloned = False
 
     def hasCommit(self, sha):
         repo = git.Repo(self.path)
@@ -261,6 +267,10 @@ class Repo(object):
         except AssertionError:
             repo.git.fetch(url, refspec)
 
+    def deleteRef(self, ref):
+        repo = git.Repo(self.path)
+        git.Reference.delete(repo, ref)
+
     def checkout(self, ref):
         repo = git.Repo(self.path)
         try:
@@ -385,8 +395,10 @@ class Repo(object):
             # f.newname = diff_context.b_path
             if diff_context.new_file:
                 f.oldname = 'Empty file'
+                f.old_empty = True
             if diff_context.deleted_file:
                 f.newname = 'Empty file'
+                f.new_empty = True
             files.append(f)
             if diff_context.rename_from:
                 f.oldname = diff_context.rename_from
@@ -464,6 +476,18 @@ class Repo(object):
                     continue
                 if not last_line:
                     raise Exception("Unhandled line: %s" % line)
+            if not diff_context.diff:
+                # There is no diff, possibly because this is simply a
+                # rename.  Include context lines so that comments may
+                # appear.
+                if not f.new_empty:
+                    blob = newc.tree[f.newname]
+                else:
+                    blob = oldc.tree[f.oldname]
+                f.old_lineno = 1
+                f.new_lineno = 1
+                for line in blob.data_stream.read().splitlines():
+                    f.addContextLine(line)
             f.finalize()
         return files
 
@@ -475,7 +499,10 @@ class Repo(object):
         f.new_lineno = 1
         repo = git.Repo(self.path)
         newc = repo.commit(new)
-        blob = newc.tree[path]
+        try:
+            blob = newc.tree[path]
+        except KeyError:
+            return None
         for line in blob.data_stream.read().splitlines():
             f.addContextLine(line)
         f.finalize()
