@@ -60,13 +60,16 @@ DEFAULT_PALETTE={
     'focused-revision-comments': ['default,standout', ''],
     'focused-revision-drafts': ['dark red,standout', ''],
     'change-message-name': ['yellow', ''],
+    'change-message-own-name': ['light cyan', ''],
     'change-message-header': ['brown', ''],
+    'change-message-own-header': ['dark cyan', ''],
     'change-message-draft': ['dark red', ''],
     'revision-button': ['dark magenta', ''],
     'focused-revision-button': ['light magenta', ''],
     'lines-added': ['light green', ''],
     'lines-removed': ['light red', ''],
     'reviewer-name': ['yellow', ''],
+    'reviewer-own-name': ['light cyan', ''],
     # project list
     'unreviewed-project': ['white', ''],
     'subscribed-project': ['default', ''],
@@ -79,6 +82,12 @@ DEFAULT_PALETTE={
     'reviewed-change': ['dark gray', ''],
     'focused-unreviewed-change': ['default,standout', ''],
     'focused-reviewed-change': ['dark gray,standout', ''],
+    'starred-change': ['light cyan', ''],
+    'focused-starred-change': ['light cyan,standout', ''],
+    'held-change': ['light red', ''],
+    'focused-held-change': ['light red,standout', ''],
+    'marked-change': ['dark cyan', ''],
+    'focused-marked-change': ['dark cyan,standout', ''],
     }
 
 # A delta from the default palette
@@ -93,8 +102,11 @@ LIGHT_PALETTE = {
     'change-data': ['dark blue,bold', ''],
     'focused-change-data': ['dark blue,standout', ''],
     'reviewer-name': ['brown', ''],
+    'reviewer-own-name': ['dark blue,bold', ''],
     'change-message-name': ['brown', ''],
+    'change-message-own-name': ['dark blue,bold', ''],
     'change-message-header': ['black', ''],
+    'change-message-own-header': ['black,bold', ''],
     'focused-link': ['dark blue,bold', ''],
     'filename': ['dark cyan', ''],
     }
