@@ -14,15 +14,17 @@
 # under the License.
 
 import datetime
+import logging
+import urlparse
 
 import urwid
 
-from gertty import gitrepo
 from gertty import keymap
 from gertty import mywid
 from gertty import sync
 from gertty.view import side_diff as view_side_diff
 from gertty.view import unified_diff as view_unified_diff
+from gertty.view import mouse_scroll_decorator
 import gertty.view
 
 class EditTopicDialog(mywid.ButtonDialog):
@@ -82,10 +84,9 @@ class CherryPickDialog(urwid.WidgetWrap):
 
 class ReviewDialog(urwid.WidgetWrap):
     signals = ['submit', 'save', 'cancel']
-    def __init__(self, revision_row):
-        self.revision_row = revision_row
-        self.change_view = revision_row.change_view
-        self.app = self.change_view.app
+    def __init__(self, app, revision_key):
+        self.revision_key = revision_key
+        self.app = app
         save_button = mywid.FixedButton(u'Save')
         submit_button = mywid.FixedButton(u'Save and Submit')
         cancel_button = mywid.FixedButton(u'Cancel')
@@ -96,20 +97,30 @@ class ReviewDialog(urwid.WidgetWrap):
         urwid.connect_signal(cancel_button, 'click',
             lambda button:self._emit('cancel'))
 
-        buttons = [('pack', save_button)]
-        if revision_row.can_submit:
-            buttons.append(('pack', submit_button))
-        buttons.append(('pack', cancel_button))
-        buttons = urwid.Columns(buttons, dividechars=2)
         rows = []
         categories = []
         values = {}
+        descriptions = {}
         self.button_groups = {}
         message = ''
         with self.app.db.getSession() as session:
-            revision = session.getRevision(self.revision_row.revision_key)
+            revision = session.getRevision(self.revision_key)
             change = revision.change
+            buttons = [('pack', save_button)]
+            if revision.can_submit:
+                buttons.append(('pack', submit_button))
+            buttons.append(('pack', cancel_button))
+            buttons = urwid.Columns(buttons, dividechars=2)
             if revision == change.revisions[-1]:
+                for label in change.labels:
+                    d = descriptions.setdefault(label.category, {})
+                    d[label.value] = label.description
+                    vmin = d.setdefault('min', label.value)
+                    if label.value < vmin:
+                        d['min'] = label.value
+                    vmax = d.setdefault('max', label.value)
+                    if label.value > vmax:
+                        d['max'] = label.value
                 for label in change.permitted_labels:
                     if label.category not in categories:
                         categories.append(label.category)
@@ -127,14 +138,26 @@ class ReviewDialog(urwid.WidgetWrap):
                         current = 0
                     else:
                         current = current.value
-                    for value in values[category]:
+                    for value in sorted(values[category], reverse=True):
                         if value > 0:
                             strvalue = '+%s' % value
                         elif value == 0:
                             strvalue = ' 0'
                         else:
                             strvalue = str(value)
+                        strvalue += '  ' + descriptions[category][value]
                         b = urwid.RadioButton(group, strvalue, state=(value == current))
+                        b._value = value
+                        if value > 0:
+                            if value == descriptions[category]['max']:
+                                b = urwid.AttrMap(b, 'max-label')
+                            else:
+                                b = urwid.AttrMap(b, 'positive-label')
+                        elif value < 0:
+                            if value == descriptions[category]['min']:
+                                b = urwid.AttrMap(b, 'min-label')
+                            else:
+                                b = urwid.AttrMap(b, 'negative-label')
                         rows.append(b)
                     rows.append(urwid.Divider())
             m = revision.getPendingMessage()
@@ -150,15 +173,14 @@ class ReviewDialog(urwid.WidgetWrap):
         fill = urwid.Filler(pile, valign='top')
         super(ReviewDialog, self).__init__(urwid.LineBox(fill, 'Review'))
 
-    def save(self, upload=False, submit=False):
+    def getValues(self):
         approvals = {}
         for category, group in self.button_groups.items():
             for button in group:
                 if button.state:
-                    approvals[category] = int(button.get_label())
+                    approvals[category] = button._value
         message = self.message.edit_text.strip()
-        self.change_view.saveReview(self.revision_row.revision_key, approvals,
-                                    message, upload, submit)
+        return (approvals, message)
 
     def keypress(self, size, key):
         r = super(ReviewDialog, self).keypress(size, key)
@@ -177,7 +199,8 @@ class ReviewButton(mywid.FixedButton):
             lambda button: self.openReview())
 
     def openReview(self):
-        self.dialog = ReviewDialog(self.revision_row)
+        self.dialog = ReviewDialog(self.change_view.app,
+                                   self.revision_row.revision_key)
         urwid.connect_signal(self.dialog, 'save',
             lambda button: self.closeReview(True, False))
         urwid.connect_signal(self.dialog, 'submit',
@@ -189,7 +212,9 @@ class ReviewButton(mywid.FixedButton):
                                    min_width=60, min_height=20)
 
     def closeReview(self, upload, submit):
-        self.dialog.save(upload, submit)
+        approvals, message = self.dialog.getValues()
+        self.change_view.saveReview(self.revision_row.revision_key, approvals,
+                                    message, upload, submit)
         self.change_view.app.backScreen()
 
 class RevisionRow(urwid.WidgetWrap):
@@ -209,22 +234,17 @@ class RevisionRow(urwid.WidgetWrap):
         self.commit_sha = revision.commit
         self.can_submit = revision.can_submit
         self.title = mywid.TextButton(u'', on_press = self.expandContract)
-        stats = repo.diffstat(revision.parent, revision.commit)
         table = mywid.Table(columns=3)
         total_added = 0
         total_removed = 0
-        for added, removed, filename in stats:
-            try:
-                added = int(added)
-            except ValueError:
-                added = 0
-            try:
-                removed = int(removed)
-            except ValueError:
-                removed = 0
+        for rfile in revision.files:
+            if rfile.status is None:
+                continue
+            added = rfile.inserted or 0
+            removed = rfile.deleted or 0
             total_added += added
             total_removed += removed
-            table.addRow([urwid.Text(('filename', filename), wrap='clip'),
+            table.addRow([urwid.Text(('filename', rfile.display_path), wrap='clip'),
                           urwid.Text([('lines-added', '+%i' % (added,)), ', '],
                                      align=urwid.RIGHT),
                           urwid.Text(('lines-removed', '-%i' % (removed,)))])
@@ -262,13 +282,13 @@ class RevisionRow(urwid.WidgetWrap):
     def update(self, revision):
         line = [('revision-name', 'Patch Set %s ' % revision.number),
                 ('revision-commit', revision.commit)]
-        num_drafts = len(revision.draft_comments)
+        num_drafts = sum([len(f.draft_comments) for f in revision.files])
         if num_drafts:
             pending_message = revision.getPendingMessage()
             if not pending_message:
                 line.append(('revision-drafts', ' (%s draft%s)' % (
                             num_drafts, num_drafts>1 and 's' or '')))
-        num_comments = len(revision.comments) - num_drafts
+        num_comments = sum([len(f.comments) for f in revision.files]) - num_drafts
         if num_comments:
             line.append(('revision-comments', ' (%s inline comment%s)' % (
                         num_comments, num_comments>1 and 's' or '')))
@@ -286,30 +306,10 @@ class RevisionRow(urwid.WidgetWrap):
         self.change_view.diff(self.revision_key)
 
     def checkout(self, button):
-        repo = self.app.getRepo(self.project_name)
-        try:
-            repo.checkout(self.commit_sha)
-            dialog = mywid.MessageDialog('Checkout', 'Change checked out in %s' % repo.path)
-            min_height=8
-        except gitrepo.GitCheckoutError as e:
-            dialog = mywid.MessageDialog('Error', e.msg)
-            min_height=12
-        urwid.connect_signal(dialog, 'close',
-            lambda button: self.app.backScreen())
-        self.app.popup(dialog, min_height=min_height)
+        self.app.localCheckoutCommit(self.project_name, self.commit_sha)
 
     def cherryPick(self, button):
-        repo = self.app.getRepo(self.project_name)
-        try:
-            repo.cherryPick(self.commit_sha)
-            dialog = mywid.MessageDialog('Cherry-Pick', 'Change cherry-picked in %s' % repo.path)
-            min_height=8
-        except gitrepo.GitCheckoutError as e:
-            dialog = mywid.MessageDialog('Error', e.msg)
-            min_height=12
-        urwid.connect_signal(dialog, 'close',
-            lambda button: self.app.backScreen())
-        self.app.popup(dialog, min_height=min_height)
+        self.app.localCherryPickCommit(self.project_name, self.commit_sha)
 
 class ChangeButton(mywid.FixedButton):
     button_left = urwid.Text(u' ')
@@ -327,7 +327,10 @@ class ChangeButton(mywid.FixedButton):
         super(ChangeButton, self).set_label(text)
 
     def openChange(self):
-        self.change_view.app.changeScreen(ChangeView(self.change_view.app, self.change_key))
+        try:
+            self.change_view.app.changeScreen(ChangeView(self.change_view.app, self.change_key))
+        except gertty.view.DisplayError as e:
+            self.change_view.app.error(e.message)
 
 class ChangeMessageBox(mywid.HyperText):
     def __init__(self, app, message):
@@ -337,14 +340,21 @@ class ChangeMessageBox(mywid.HyperText):
 
     def refresh(self, message):
         self.message_created = message.created
+        created = self.app.time(message.created)
         lines = message.message.split('\n')
         if message.draft:
             lines.insert(0, '')
             lines.insert(0, 'Patch Set %s:' % (message.revision.number,))
-        text = [('change-message-name', message.author_name),
-                ('change-message-header', ': '+lines.pop(0)),
-                ('change-message-header',
-                 message.created.strftime(' (%Y-%m-%d %H:%M:%S%z)'))]
+        if message.author.username == self.app.config.username:
+            name_style = 'change-message-own-name'
+            header_style = 'change-message-own-header'
+        else:
+            name_style = 'change-message-name'
+            header_style = 'change-message-header'
+        text = [(name_style, message.author_name),
+                (header_style, ': '+lines.pop(0)),
+                (header_style,
+                 created.strftime(' (%Y-%m-%d %H:%M:%S%z)'))]
         if message.draft and not message.pending:
             text.append(('change-message-draft', ' (draft)'))
         if lines and lines[-1]:
@@ -365,6 +375,7 @@ class CommitMessageBox(mywid.HyperText):
             text = commentlink.run(self.app, text)
         super(CommitMessageBox, self).set_text(text)
 
+@mouse_scroll_decorator.ScrollByWheel
 class ChangeView(urwid.WidgetWrap):
     def help(self):
         key = self.app.config.keymap.formatKeys
@@ -372,7 +383,7 @@ class ChangeView(urwid.WidgetWrap):
             (key(keymap.LOCAL_CHECKOUT),
              "Checkout the most recent revision into the local repo"),
             (key(keymap.DIFF),
-             "Show the diff of the mont recent revision"),
+             "Show the diff of the most recent revision"),
             (key(keymap.TOGGLE_HIDDEN),
              "Toggle the hidden flag for the current change"),
             (key(keymap.NEXT_CHANGE),
@@ -381,12 +392,16 @@ class ChangeView(urwid.WidgetWrap):
              "Go to the previous change in the list"),
             (key(keymap.REVIEW),
              "Leave a review for the most recent revision"),
+            (key(keymap.TOGGLE_HELD),
+             "Toggle the held flag for the current change"),
             (key(keymap.TOGGLE_HIDDEN_COMMENTS),
              "Toggle display of hidden comments"),
             (key(keymap.SEARCH_RESULTS),
              "Back to the list of changes"),
             (key(keymap.TOGGLE_REVIEWED),
              "Toggle the reviewed flag for the current change"),
+            (key(keymap.TOGGLE_STARRED),
+             "Toggle the starred flag for the current change"),
             (key(keymap.LOCAL_CHERRY_PICK),
              "Cherry-pick the most recent revision onto the local repo"),
             (key(keymap.ABANDON_CHANGE),
@@ -415,29 +430,41 @@ class ChangeView(urwid.WidgetWrap):
 
     def __init__(self, app, change_key):
         super(ChangeView, self).__init__(urwid.Pile([]))
+        self.log = logging.getLogger('gertty.view.change')
         self.app = app
         self.change_key = change_key
         self.revision_rows = {}
         self.message_rows = {}
         self.last_revision_key = None
         self.hide_comments = True
-        self.change_id_label = urwid.Text(u'', wrap='clip')
-        self.owner_label = urwid.Text(u'', wrap='clip')
-        self.project_label = urwid.Text(u'', wrap='clip')
+        self.change_id_label = mywid.TextButton(u'', on_press=self.searchChangeId)
+        self.owner_label = mywid.TextButton(u'', on_press=self.searchOwner)
+        self.project_label = mywid.TextButton(u'', on_press=self.searchProject)
         self.branch_label = urwid.Text(u'', wrap='clip')
-        self.topic_label = urwid.Text(u'', wrap='clip')
+        self.topic_label = mywid.TextButton(u'', on_press=self.searchTopic)
         self.created_label = urwid.Text(u'', wrap='clip')
         self.updated_label = urwid.Text(u'', wrap='clip')
         self.status_label = urwid.Text(u'', wrap='clip')
+        self.permalink_label = urwid.Text(u'', wrap='clip')
         change_info = []
-        for l, v in [("Change-Id", self.change_id_label),
-                     ("Owner", self.owner_label),
-                     ("Project", self.project_label),
+        change_info_map={'change-data': 'focused-change-data'}
+        for l, v in [("Change-Id", urwid.Padding(urwid.AttrMap(self.change_id_label, None,
+                                                               focus_map=change_info_map),
+                                                 width='pack')),
+                     ("Owner", urwid.Padding(urwid.AttrMap(self.owner_label, None,
+                                                           focus_map=change_info_map),
+                                             width='pack')),
+                     ("Project", urwid.Padding(urwid.AttrMap(self.project_label, None,
+                                                           focus_map=change_info_map),
+                                             width='pack')),
                      ("Branch", self.branch_label),
-                     ("Topic", self.topic_label),
+                     ("Topic", urwid.Padding(urwid.AttrMap(self.topic_label, None,
+                                                           focus_map=change_info_map),
+                                             width='pack')),
                      ("Created", self.created_label),
                      ("Updated", self.updated_label),
                      ("Status", self.status_label),
+                     ("Permalink", self.permalink_label),
                      ]:
             row = urwid.Columns([(12, urwid.Text(('change-header', l), wrap='clip')), v])
             change_info.append(row)
@@ -497,44 +524,58 @@ class ChangeView(urwid.WidgetWrap):
             if not succeeded:
                 raise gertty.view.DisplayError("Git commits not present in local repository")
 
+    def interested(self, event):
+        if not ((isinstance(event, sync.ChangeAddedEvent) and
+                 self.change_key in event.related_change_keys)
+                or
+                (isinstance(event, sync.ChangeUpdatedEvent) and
+                 self.change_key in event.related_change_keys)):
+            self.log.debug("Ignoring refresh change due to event %s" % (event,))
+            return False
+        self.log.debug("Refreshing change due to event %s" % (event,))
+        return True
+
     def refresh(self):
-        change_info = []
         with self.app.db.getSession() as session:
             change = session.getChange(self.change_key)
             self.topic = change.topic or ''
             self.pending_status_message = change.pending_status_message or ''
+            reviewed = hidden = starred = held = ''
             if change.reviewed:
                 reviewed = ' (reviewed)'
-            else:
-                reviewed = ''
             if change.hidden:
                 hidden = ' (hidden)'
-            else:
-                hidden = ''
-            self.title = 'Change %s%s%s' % (change.number, reviewed, hidden)
+            if change.starred:
+                starred = '* '
+            if change.held:
+                held = ' (held)'
+            self.title = '%sChange %s%s%s%s' % (starred, change.number, reviewed,
+                                                hidden, held)
             self.app.status.update(title=self.title)
             self.project_key = change.project.key
+            self.project_name = change.project.name
             self.change_rest_id = change.id
+            self.change_id = change.change_id
+            if change.owner:
+                self.owner_email = change.owner.email
+            else:
+                self.owner_email = None
 
-            self.change_id_label.set_text(('change-data', change.change_id))
-            self.owner_label.set_text(('change-data', change.owner_name))
-            self.project_label.set_text(('change-data', change.project.name))
+            self.change_id_label.text.set_text(('change-data', change.change_id))
+            self.owner_label.text.set_text(('change-data', change.owner_name))
+            self.project_label.text.set_text(('change-data', change.project.name))
             self.branch_label.set_text(('change-data', change.branch))
-            self.topic_label.set_text(('change-data', self.topic))
-            self.created_label.set_text(('change-data', str(change.created)))
-            self.updated_label.set_text(('change-data', str(change.updated)))
+            self.topic_label.text.set_text(('change-data', self.topic))
+            self.created_label.set_text(('change-data', str(self.app.time(change.created))))
+            self.updated_label.set_text(('change-data', str(self.app.time(change.updated))))
             self.status_label.set_text(('change-data', change.status))
+            url = urlparse.urljoin(self.app.config.url, str(change.number))
+            self.permalink_label.set_text(('change-data', url))
             self.commit_message.set_text(change.revisions[-1].message)
 
             categories = []
-            max_values = {}
-            min_values = {}
             approval_headers = [urwid.Text(('table-header', 'Name'))]
             for label in change.labels:
-                if label.value > max_values.get(label.category, 0):
-                    max_values[label.category] = label.value
-                if label.value < min_values.get(label.category, 0):
-                    min_values[label.category] = label.value
                 if label.category in categories:
                     continue
                 approval_headers.append(urwid.Text(('table-header', label.category)))
@@ -550,7 +591,11 @@ class ChangeView(urwid.WidgetWrap):
                 if not approvals:
                     approvals = {}
                     row = []
-                    row.append(urwid.Text(('reviewer-name', approval.reviewer.name)))
+                    if approval.reviewer.username == self.app.config.username:
+                        style = 'reviewer-own-name'
+                    else:
+                        style = 'reviewer-name'
+                    row.append(urwid.Text((style, approval.reviewer.name)))
                     for i, category in enumerate(categories):
                         w = urwid.Text(u'', align=urwid.CENTER)
                         approvals[category] = w
@@ -558,15 +603,16 @@ class ChangeView(urwid.WidgetWrap):
                     approvals_for_name[approval.reviewer.name] = approvals
                     votes.addRow(row)
                 if str(approval.value) != '0':
+                    cat_min, cat_max = change.getMinMaxPermittedForCategory(approval.category)
                     if approval.value > 0:
                         val = '+%i' % approval.value
-                        if approval.value == max_values.get(approval.category):
+                        if approval.value == cat_max:
                             val = ('max-label', val)
                         else:
                             val = ('positive-label', val)
                     else:
                         val = '%i' % approval.value
-                        if approval.value == min_values.get(approval.category):
+                        if approval.value == cat_min:
                             val = ('min-label', val)
                         else:
                             val = ('negative-label', val)
@@ -703,11 +749,10 @@ class ChangeView(urwid.WidgetWrap):
 
         # Handle needed-by
         children = {}
-        for revision in change.revisions:
-            children.update((r.change.key, r.change.subject)
-                            for r in session.getRevisionsByParent(revision.commit)
-                            if (r.change.status != 'MERGED' and
-                                r == r.change.revisions[-1]))
+        children.update((r.change.key, r.change.subject)
+                        for r in session.getRevisionsByParent([revision.commit for revision in change.revisions])
+                        if (r.change.status != 'MERGED' and
+                            r == r.change.revisions[-1]))
         self._updateDependenciesWidget(children,
                                        self.needed_by, self.needed_by_rows,
                                        header='Needed by:')
@@ -723,6 +768,17 @@ class ChangeView(urwid.WidgetWrap):
             change = session.getChange(self.change_key)
             change.hidden = not change.hidden
 
+    def toggleStarred(self):
+        with self.app.db.getSession() as session:
+            change = session.getChange(self.change_key)
+            change.starred = not change.starred
+            change.pending_starred = True
+        self.app.sync.submitTask(
+            sync.ChangeStarredTask(self.change_key, sync.HIGH_PRIORITY))
+
+    def toggleHeld(self):
+        return self.app.toggleHeldChange(self.change_key)
+
     def keypress(self, size, key):
         r = super(ChangeView, self).keypress(size, key)
         commands = self.app.config.keymap.getCommands(r)
@@ -734,6 +790,14 @@ class ChangeView(urwid.WidgetWrap):
             self.toggleHidden()
             self.refresh()
             return None
+        if keymap.TOGGLE_STARRED in commands:
+            self.toggleStarred()
+            self.refresh()
+            return None
+        if keymap.TOGGLE_HELD in commands:
+            self.toggleHeld()
+            self.refresh()
+            return None
         if keymap.REVIEW in commands:
             row = self.revision_rows[self.last_revision_key]
             row.review_button.openReview()
@@ -949,6 +1013,20 @@ class ChangeView(urwid.WidgetWrap):
         self.app.backScreen()
         self.refresh()
 
+    def searchChangeId(self, widget):
+        self.app.doSearch("status:open change:%s" % (self.change_id,))
+
+    def searchOwner(self, widget):
+        if self.owner_email:
+            self.app.doSearch("status:open owner:%s" % (self.owner_email,))
+
+    def searchProject(self, widget):
+        self.app.doSearch("status:open project:%s" % (self.project_name,))
+
+    def searchTopic(self, widget):
+        if self.topic:
+            self.app.doSearch("status:open topic:%s" % (self.topic,))
+
     def reviewKey(self, reviewkey):
         approvals = {}
         for a in reviewkey['approvals']:
@@ -960,46 +1038,10 @@ class ChangeView(urwid.WidgetWrap):
         self.saveReview(row.revision_key, approvals, '', True, submit)
 
     def saveReview(self, revision_key, approvals, message, upload, submit):
-        message_key = None
-        with self.app.db.getSession() as session:
-            account = session.getAccountByUsername(self.app.config.username)
-            revision = session.getRevision(revision_key)
-            change = revision.change
-            draft_approvals = {}
-            for approval in change.draft_approvals:
-                draft_approvals[approval.category] = approval
-
-            categories = set()
-            for label in change.permitted_labels:
-                categories.add(label.category)
-            for category in categories:
-                value = approvals.get(category, 0)
-                approval = draft_approvals.get(category)
-                if not approval:
-                    approval = change.createApproval(account, category, 0, draft=True)
-                    draft_approvals[category] = approval
-                approval.value = value
-            draft_message = revision.getPendingMessage()
-            if not draft_message:
-                draft_message = revision.getDraftMessage()
-            if not draft_message:
-                if message or upload:
-                    draft_message = revision.createMessage(None, account,
-                                                           datetime.datetime.utcnow(),
-                                                           '', draft=True)
-            if draft_message:
-                draft_message.created = datetime.datetime.utcnow()
-                draft_message.message = message
-                draft_message.pending = upload
-                message_key = draft_message.key
-            if upload:
-                change.reviewed = True
-            if submit:
-                change.status = 'SUBMITTED'
-                change.pending_status = True
-                change.pending_status_message = None
-        # Outside of db session
+        message_keys = self.app.saveReviews([revision_key], approvals,
+                                            message, upload, submit)
         if upload:
-            self.app.sync.submitTask(
-                sync.UploadReviewTask(message_key, sync.HIGH_PRIORITY))
+            for message_key in message_keys:
+                self.app.sync.submitTask(
+                    sync.UploadReviewTask(message_key, sync.HIGH_PRIORITY))
         self.refresh()
