@@ -17,9 +17,9 @@ import logging
 import urwid
 
 from gertty import keymap
-from gertty import mywid
 from gertty import sync
 from gertty.view import change_list as view_change_list
+from gertty.view import mouse_scroll_decorator
 
 class ProjectRow(urwid.Button):
     project_focus_map = {None: 'focused',
@@ -67,6 +67,7 @@ class ProjectListHeader(urwid.WidgetWrap):
                 (5, urwid.Text(u'Open'))]
         super(ProjectListHeader, self).__init__(urwid.Columns(cols))
 
+@mouse_scroll_decorator.ScrollByWheel
 class ProjectListView(urwid.WidgetWrap):
     def help(self):
         key = self.app.config.keymap.formatKeys
@@ -78,7 +79,7 @@ class ProjectListView(urwid.WidgetWrap):
             (key(keymap.TOGGLE_SUBSCRIBED),
              "Toggle the subscription flag for the currently selected project"),
             (key(keymap.REFRESH),
-             "Sync all projects")
+             "Sync subscribed projects")
             ]
 
     def __init__(self, app):
@@ -148,7 +149,7 @@ class ProjectListView(urwid.WidgetWrap):
         self.app.changeScreen(view_change_list.ChangeListView(
                 self.app,
                 "_project_key:%s %s" % (project_key, self.app.config.project_change_list_query),
-                project_name, project_key=None, unreviewed=True))
+                project_name, project_key=project_key, unreviewed=True))
 
     def keypress(self, size, key):
         r = super(ProjectListView, self).keypress(size, key)
