@@ -12,15 +12,13 @@
 # License for the specific language governing permissions and limitations
 # under the License.
 
-import datetime
-import logging
-
 import urwid
 
 from gertty import keymap
-from gertty import mywid
-from gertty import gitrepo
-from gertty.view.diff import *
+from gertty.view.diff import BaseDiffComment, BaseDiffCommentEdit, BaseDiffLine
+from gertty.view.diff import BaseFileHeader, BaseFileReminder, BaseDiffView
+
+LN_COL_WIDTH = 5
 
 class SideDiffCommentEdit(BaseDiffCommentEdit):
     def __init__(self, app, context, old_key=None, new_key=None, old=u'', new=u''):
@@ -32,21 +30,35 @@ class SideDiffCommentEdit(BaseDiffCommentEdit):
         self.new_key = new_key
         self.old = urwid.Edit(edit_text=old, multiline=True)
         self.new = urwid.Edit(edit_text=new, multiline=True)
-        self.contents.append((urwid.Text(u''), ('given', 4, False)))
-        self.contents.append((urwid.AttrMap(self.old, 'draft-comment'), ('weight', 1, False)))
-        self.contents.append((urwid.Text(u''), ('given', 4, False)))
-        self.contents.append((urwid.AttrMap(self.new, 'draft-comment'), ('weight', 1, False)))
-        self.focus_position = 3
+        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
+        if context.old_file_key and (context.old_ln is not None or context.header):
+            self.contents.append((urwid.AttrMap(self.old, 'draft-comment'), ('weight', 1, False)))
+        else:
+            self.contents.append((urwid.Text(u''), ('weight', 1, False)))
+        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
+        if context.new_file_key and (context.new_ln is not None or context.header):
+            self.contents.append((urwid.AttrMap(self.new, 'draft-comment'), ('weight', 1, False)))
+            new_editable = True
+        else:
+            self.contents.append((urwid.Text(u''), ('weight', 1, False)))
+            new_editable = False
+        if new_editable:
+            self.focus_position = 3
+        else:
+            self.focus_position = 1
 
     def keypress(self, size, key):
         r = super(SideDiffCommentEdit, self).keypress(size, key)
         commands = self.app.config.keymap.getCommands(r)
         if ((keymap.NEXT_SELECTABLE in commands) or
             (keymap.PREV_SELECTABLE in commands)):
-            if self.focus_position == 3:
-                self.focus_position = 1
-            else:
-                self.focus_position = 3
+            if ((self.context.old_ln is not None and
+                 self.context.new_ln is not None) or
+                self.context.header):
+                if self.focus_position == 3:
+                    self.focus_position = 1
+                else:
+                    self.focus_position = 3
             return None
         return r
 
@@ -60,9 +72,9 @@ class SideDiffComment(BaseDiffComment):
             oldt = urwid.AttrMap(oldt, 'comment')
         if new:
             newt = urwid.AttrMap(newt, 'comment')
-        self.contents.append((urwid.Text(u''), ('given', 4, False)))
+        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
         self.contents.append((oldt, ('weight', 1, False)))
-        self.contents.append((urwid.Text(u''), ('given', 4, False)))
+        self.contents.append((urwid.Text(u''), ('given', LN_COL_WIDTH, False)))
         self.contents.append((newt, ('weight', 1, False)))
 
 class SideDiffLine(BaseDiffLine):
@@ -74,13 +86,13 @@ class SideDiffLine(BaseDiffLine):
             if ln is None:
                 ln = ''
             else:
-                ln = str(ln)
+                ln = '%*i' % (LN_COL_WIDTH-1, ln)
             ln_col = urwid.Text(('line-number', ln))
             ln_col.set_wrap_mode('clip')
             line_col = urwid.Text(line)
             if action == '':
                 line_col = urwid.AttrMap(line_col, 'nonexistent')
-            columns += [(4, ln_col), line_col]
+            columns += [(LN_COL_WIDTH, ln_col), line_col]
         col = urwid.Columns(columns)
         map = {None: 'focused',
                'added-line': 'focused-added-line',
@@ -103,16 +115,22 @@ class SideFileHeader(BaseFileHeader):
                'filename': 'focused-filename'}
         self._w = urwid.AttrMap(col, None, focus_map=map)
 
+class SideFileReminder(BaseFileReminder):
+    def __init__(self):
+        self.old_text = urwid.Text(('filename', ''))
+        self.new_text = urwid.Text(('filename', ''))
+        col = urwid.Columns([self.old_text, self.new_text])
+        super(SideFileReminder, self).__init__(col)
+
+    def set(self, old, new):
+        self.old_text.set_text(('filename', old))
+        self.new_text.set_text(('filename', new))
 
 class SideDiffView(BaseDiffView):
     def makeLines(self, diff, lines_to_add, comment_lists):
         lines = []
         for old, new in lines_to_add:
-            context = LineContext(
-                self.old_revision_key, self.new_revision_key,
-                self.old_revision_num, self.new_revision_num,
-                diff.oldname, diff.newname,
-                old[0], new[0])
+            context = self.makeContext(diff, old[0], new[0])
             lines.append(SideDiffLine(self.app, context, old, new,
                                       callback=self.onSelect))
             # see if there are any comments for this line
@@ -146,12 +164,11 @@ class SideDiffView(BaseDiffView):
                                                  old_comment, new_comment))
         return lines
 
+    def makeFileReminder(self):
+        return SideFileReminder()
+
     def makeFileHeader(self, diff, comment_lists):
-        context = LineContext(
-            self.old_revision_key, self.new_revision_key,
-            self.old_revision_num, self.new_revision_num,
-            diff.oldname, diff.newname,
-            None, None)
+        context = self.makeContext(diff, None, None, header=True)
         lines = []
         lines.append(SideFileHeader(self.app, context, diff.oldname, diff.newname,
                                     callback=self.onSelect))
